/*
 * Copyright 2019 ConsenSys AG.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 */
package tech.pegasys.pantheon.ethereum.eth.sync.fullsync;

import tech.pegasys.pantheon.ethereum.eth.manager.ChainState;
import tech.pegasys.pantheon.ethereum.eth.manager.EthPeer;
import tech.pegasys.pantheon.ethereum.eth.manager.EthPeers;
import tech.pegasys.pantheon.ethereum.eth.sync.SynchronizerConfiguration;
import tech.pegasys.pantheon.util.uint.UInt256;

import java.util.Optional;

public class BetterSyncTargetEvaluator {

  private final SynchronizerConfiguration config;
  private final EthPeers ethPeers;

  public BetterSyncTargetEvaluator(
      final SynchronizerConfiguration config, final EthPeers ethPeers) {
    this.config = config;
    this.ethPeers = ethPeers;
  }

  public boolean shouldSwitchSyncTarget(final EthPeer currentSyncTarget) {
    final ChainState currentPeerChainState = currentSyncTarget.chainState();
    final Optional<EthPeer> maybeBestPeer = ethPeers.bestPeer();

    return maybeBestPeer
        .map(
            bestPeer -> {
              if (EthPeers.BEST_CHAIN.compare(bestPeer, currentSyncTarget) <= 0) {
                // Our current target is better or equal to the best peer
                return false;
              }
              // Require some threshold to be exceeded before switching targets to keep some
              // stability when multiple peers are in range of each other
              final ChainState bestPeerChainState = bestPeer.chainState();
              final UInt256 tdDifference =
                  bestPeerChainState
                      .getBestBlock()
                      .getTotalDifficulty()
                      .minus(currentPeerChainState.getBestBlock().getTotalDifficulty());
              if (tdDifference.compareTo(config.downloaderChangeTargetThresholdByTd()) > 0) {
                return true;
              }
              final long heightDifference =
                  bestPeerChainState.getEstimatedHeight()
                      - currentPeerChainState.getEstimatedHeight();
              return heightDifference > config.downloaderChangeTargetThresholdByHeight();
            })
        .orElse(false);
  }
}
